# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/jerx.prompt.ipynb.

# %% auto 0
__all__ = ['log', 'DEFAULT_SYSTEM_PROMPT_TEMPLATE', 'DEFAULT_RELATION_SET_PROMPT_TEMPLATE', 'JERXChatFormatter']

# %% ../../nbs/jerx.prompt.ipynb 3
import random
from dataclasses import dataclass
from typing import Iterable, Callable, Any, Generator

from ..logging import get_logger

log = get_logger(__name__)

# %% ../../nbs/jerx.prompt.ipynb 6
DEFAULT_SYSTEM_PROMPT_TEMPLATE = """You are a helpful assistant that extracts up to {max_triplets} entity-relation-entity triplets from given text. Use '|' as delimiter and provide one triplet per line. The entities in a triplet must be different.
{relation_set_prompt}
""".strip()

DEFAULT_RELATION_SET_PROMPT_TEMPLATE = """Here are the list of relations that you can use:
{relation_set}
""".strip()

@dataclass
class JERXChatFormatter:
    system_prompt_template: str = DEFAULT_SYSTEM_PROMPT_TEMPLATE
    relation_set_prompt_template: str = DEFAULT_RELATION_SET_PROMPT_TEMPLATE
    relation_set: set|None = None
    few_shot_examples: list[dict]|None = None
    n_few_shot_examples: int = 3

    def __post_init__(self):
        if self.relation_set:
            self.relation_set = sorted(self.relation_set)
        if self.few_shot_examples is not None:
            self.few_shot_examples = list(self.few_shot_examples)

    def format(self, example: dict):
        few_shot_examples = self._choose_few_shot_examples()
        max_triplets = max(*[len(example['triplets']) for example in few_shot_examples], len(example['triplets'])) + 2
        messages = [
            {"role": "system", "content": self.make_system_message(max_triplets=max_triplets)},
            *list(self.make_messages(*few_shot_examples, example)),
        ]
        return {'conversations': messages}

    def make_system_message(self, max_triplets: int = 5) -> str:
        rsp = self.relation_set_prompt_template.format(relation_set=','.join(self.relation_set)) if self.relation_set else ""
        return self.system_prompt_template.format(max_triplets=max_triplets, relation_set_prompt=rsp)

    def make_messages(self, *examples) -> Generator[dict, None, None]:
        for example in examples:
            yield {"role": "user", "content": example["text"]}
            yield {"role": "assistant", "content": self._format_triplets(example["triplets"])}

    def _choose_few_shot_examples(self) -> list[dict]:
        if len(self.few_shot_examples) <= self.n_few_shot_examples:
            return self.few_shot_examples
        else:
            return random.sample(self.few_shot_examples, k=self.n_few_shot_examples)

    def _format_triplets(self, triplets: Iterable[str]) -> str:
        return '\n'.join(sorted(triplets))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/langchain.obs.ipynb.

# %% auto 0
__all__ = ['patch_wandb_tracer_serialize_io']

# %% ../../nbs/langchain.obs.ipynb 3
from typing import Optional

# %% ../../nbs/langchain.obs.ipynb 4
def _serialize_io(run_inputs: Optional[dict]) -> dict:
    """Different to the original function, this also handles the serialization of Pydantic output from llms."""
    if not run_inputs:
        return {}
    from google.protobuf.json_format import MessageToJson
    from google.protobuf.message import Message
    from langchain_core.pydantic_v1 import BaseModel

    serialized_inputs = {}
    for key, value in run_inputs.items():
        if isinstance(value, Message):
            serialized_inputs[key] = MessageToJson(value)
        elif key == "input_documents":
            serialized_inputs.update(
                {f"input_document_{i}": doc.json() for i, doc in enumerate(value)}
            )
        elif isinstance(value, BaseModel): # this is added
            serialized_inputs[key] = value.json()
        else:
            serialized_inputs[key] = value
    return serialized_inputs

def patch_wandb_tracer_serialize_io():
    from langchain.callbacks.tracers import wandb
    wandb._serialize_io = _serialize_io



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/text.utils.ipynb.

# %% auto 0
__all__ = ['similarity', 'fuzzy_match']

# %% ../../nbs/text.utils.ipynb 3
from difflib import SequenceMatcher

# %% ../../nbs/text.utils.ipynb 4
def similarity(a, b):
    return SequenceMatcher(None, a, b).ratio()

def fuzzy_match(a: str, b: str, threshold: float = 0.7):
    return similarity(a, b) >= threshold



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/llama_index.obs.ipynb.

# %% auto 0
__all__ = ['TraceRecorder', 'make_phoenix_trace_callback_handler']

# %% ../../nbs/llama_index.obs.ipynb 3
from io import TextIOBase
from pathlib import Path
from phoenix.trace.llama_index import OpenInferenceTraceCallbackHandler
from phoenix.trace.schemas import Span
from phoenix.trace.span_json_encoder import spans_to_jsonl
from typing import Iterator

# %% ../../nbs/llama_index.obs.ipynb 4
class TraceRecorder:
    def __init__(self, f: TextIOBase | Path):
        if isinstance(f, Path):
            f.parent.mkdir(parents=True, exist_ok=True)
            self.f = open(f, "a")
        else:
            self.f = f

    def __call__(self, spans: Iterator[Span]) -> None:
        self.f.write("\n")
        self.f.write(spans_to_jsonl(spans))


def make_phoenix_trace_callback_handler(output_filepath: Path):
    return OpenInferenceTraceCallbackHandler(callback=TraceRecorder(output_filepath))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/hf.transformers.experiment.ipynb.

# %% auto 0
__all__ = ['log', 'preprocess_config', 'fine_tune']

# %% ../../../nbs/hf.transformers.experiment.ipynb 3
from copy import deepcopy
import torch
from ...utils import NestedDict, generate_time_id
from ...logging import get_logger
from .utils import load_tokenizer_model

log = get_logger(__name__)

# %% ../../../nbs/hf.transformers.experiment.ipynb 4
def preprocess_config(config: NestedDict):
    config = deepcopy(config)

    # Set float precision
    major, _ = torch.cuda.get_device_capability()
    if major >= 8:
        log.info("GPU supports bfloat16.")
        torch_dtype, bf16, fp16, bnb_4bit_compute_dtype = ("bfloat16", True, False, "bfloat16")
    else:
        log.info("GPU does not support bfloat16.")
        torch_dtype, bf16, fp16, bnb_4bit_compute_dtype = ("float16", False, True, "float16")

    if config.at("pretrained_model.torch_dtype"):
        config.set("pretrained_model.torch_dtype", torch_dtype)
    if config.at("pretrained_model.quantization_config.load_in_4bit"):
        config.set("pretrained_model.quantization_config.bnb_4bit_compute_dtype", bnb_4bit_compute_dtype)
    if config.at("trainer.training_args.bf16") or config.at("trainer.training_args.fp16"):
        config.set("trainer.training_args.bf16", bf16)
        config.set("trainer.training_args.fp16", fp16)
    
    # Generate unique model id
    model_id = config.at("hfhub.model_id")
    if config.at("trainer.lora") and "-peft" not in model_id:
        model_id += "-peft"
    if "debug" not in model_id:
        model_id += f"-{generate_time_id()}"
    config.set("hfhub.model_id", model_id)

    return config


# %% ../../../nbs/hf.transformers.experiment.ipynb 5
def fine_tune(config: NestedDict):
    from math import ceil
    from datasets import load_dataset
    from peft import LoraConfig
    from transformers import TrainingArguments
    from trl import DataCollatorForCompletionOnlyLM, SFTTrainer

    # Base model
    pretrained_model_config = config["pretrained_model"]
    model_id = pretrained_model_config.pop("model_name_or_path")
    tokenizer, base_model = load_tokenizer_model(model_id, **pretrained_model_config)
    log.info(f"Loaded base model {model_id}")

    if "llama" in model_id:
        from bellek.hf.transformers.llama import prepare_llama2_for_training

        log.info("Base model is a LLAMA model, preparing it for training.")
        prepare_llama2_for_training(tokenizer, base_model)

    # Train dataset
    train_ds = load_dataset(**config.at("dataset.train"))
    log.info(f"Loaded training dataset with {len(train_ds)} samples.")

    # Inspect token counts
    tokenized_train_ds = train_ds.map(lambda examples: tokenizer(examples["text"]), batched=True)
    token_counts = [len(input_ids) for input_ids in tokenized_train_ds["input_ids"]]
    log.info(f"Input token counts: min={min(token_counts)}, max={max(token_counts)}")

    # Supervised fine-tuning
    if config.at("trainer.max_seq_length") is None:
        config.set("trainer.max_seq_length", ceil(max(token_counts) / 8) * 8)
    max_seq_length = config.at("trainer.max_seq_length")
    log.info(f"Setting max_seq_length={max_seq_length}")

    packing = config.at("trainer.packing", False)

    if response_template := config.at("trainer.response_template"):
        log.info(f"Using completion-only data collator with response template '{response_template}'")
        data_collator = DataCollatorForCompletionOnlyLM(response_template, tokenizer=tokenizer)
    else:
        data_collator = None
    
    peft_config = LoraConfig(**config.at("trainer.lora", {}))
    training_args = TrainingArguments(
        output_dir="./results",
        **config.at("trainer.training_args"),
    )
    trainer = SFTTrainer(
        model=base_model,
        tokenizer=tokenizer,
        peft_config=peft_config,
        train_dataset=train_ds,
        dataset_text_field="text",
        max_seq_length=max_seq_length,
        packing=packing,
        data_collator=data_collator,
        args=training_args,
    )
    log.info("Starting training...")
    trainer.train()

    # Save trained model
    log.info("Saving model...")
    final_model_id = config.at("hfhub.model_id")
    trainer.model.push_to_hub(final_model_id)
    tokenizer.push_to_hub(final_model_id)
    log.info(f"Uploaded PEFT adapters to HF Hub as {final_model_id}")

    return trainer

